#!/usr/bin/env python
# -*- coding: utf-8 -*-

#****************************************
# Copyright 2009 JamiQ http://www.jamiq.com/
#****************************************

"""Parses HTML and extracts main text parts.

ExtMainText parses HTML document, extracts main text, and filters out
advertisements and common menus from it.

This module bases on version 0.1a of http://www.elias.cn/En/ExtMainText ,
but rewrite using lxml, and counts on the density of pure text versus total
length of original html description. Current algorithm is inspired by
http://ai-depot.com/articles/the-easy-way-to-extract-useful-text-from-arbitrary-html/

The old version (0.1a and earlier) bases on the measure of html tag
density, and determines density threshold according to historical
experience. Related algorithm comes from
http://www.xd-tech.com.cn/blog/article.asp?id=59

Attention:
    Current module can not pick up short message such as Twitter ones,
    because very small text looks the same as ads and common parts of pages.

    For sites like Twitter, analyzing html tag and CSS class structure
    is a better way to get main text.

Usage:
    There are two modes to call ExtMainText according to the 'filterMode'
    parameter:
    * When 'filterMode' is False (by default), it's the normal 'Extract'
    mode, which will try to find out and return the major content of html
    document, and filter out all rubbish, even comments relate to the
    document. And it's possible that the module eat the title of the
    document for specific page style.
    * When 'filterMode' is True, it's the 'Filter' mode, which will
    filter out all invaluable parts, such as ads, spams, menu links, etc,
    and return the remain part (Not really ALL of them, most of them
    in fact, depends on page presentation style). Normally, this mode
    is safer than the other, because it reserve as many material as it can.
    Usually, comments could be reserved in this mode.

    The "if __name__ == '__main__'" block gives out simple demo for how
    to use ExtMainText.

    The static threshold 0.5 works good enough for most web sites in
    English language. For other language, such as Chinese, you need to
    analyze the distribution of density for each html tag branch, and
    choose a threshold clearly distinguish main content and others.
"""

from lxml import etree
#_parser = etree.HTMLParser()
#parse = lambda x: etree.fromstring(x, _parser)
# SoupParser is more robust than lxml's default html parser sometimes:
import lxml.html.soupparser
parse = lxml.html.soupparser.fromstring

def extMainText(html, threshold = 0.5, filterMode = False):
    """
    Parses HTML and keeps only main text parts.

    PARAMETERS:
    threshold - The density threshold to distinguish major content & others.
    filterMode - Use normal 'Extract' mode or the other 'Filter' mode.

    RETURN:
    html fragments of main text
    """
    root = parse(html)
    densDic = _calcDensity(root)
    if filterMode:
        return _filterSpam(densDic, threshold)
    else:
        maxPart, textLen, maxPartChilds, textLenChilds = _getMainText(densDic, threshold)
        if textLenChilds > textLen:
            return ''.join(map(lambda tree: etree.tostring(tree, encoding = unicode) if tree != None else '', maxPartChilds))
        else:
            return etree.tostring(maxPart, encoding = unicode) if maxPart != None else ''

def _getMainText(densDic, threshold):
    """
    Get the largest html fragment with text density larger than threshold according 
    to density dictionary.

    And the largest html fragment could be made up of several continuous brother
    html branches.

    Return: (etree instance, text length, list of child etrees, total text length of child etrees)
    """
    dens, textLen, totalLen, tree = densDic['self']
    maxChildTrees = []
    maxChildTreesTextLen = 0
    # If density is bigger than threshold, current tag branch is the largest:
    if dens >= threshold:
        maxTree = tree
        maxTextLen = textLen
    # If density of current tag branch is too small, check its children:
    else:
        maxTree = None
        maxTextLen = 0
        maxChildSubTrees = []
        maxChildSubTreesTextLen = 0
        childTreesTmp = []
        childTreesTmpTextLens = []
        childTreesTmpTotalLens= []
        if densDic.has_key('child'):
            for childDic in densDic['child']:
                childDens, childTextLen, childTotalLen, childTree = childDic['self']
                tree, textLen, childTrees, childTreesTextLen = _getMainText(childDic, threshold)
                # Remember the largest tag branches of children:
                if childTreesTextLen > maxChildSubTreesTextLen:
                    maxChildSubTrees, maxChildSubTreesTextLen = childTrees, childTreesTextLen
                childTreesTmp.append(childTree)
                childTreesTmpTextLens.append(childTextLen)
                childTreesTmpTotalLens.append(childTotalLen)
                if textLen > maxTextLen:
                    maxTree = tree
                    maxTextLen = textLen
            # Find the largest html fragment under current tag branch:
            for j in range(1, len(childTreesTmp) + 1):
                for i in range(j):
                    childTreesTmpTotalLen= sum(childTreesTmpTotalLens[i:j])
                    childTreesTmpTextLen = sum(childTreesTmpTextLens[i:j])
                    childTreesTmpTotalLen = 1 if childTreesTmpTotalLen == 0 else childTreesTmpTotalLen
                    if float(childTreesTmpTextLen) / childTreesTmpTotalLen >= threshold:
                        if childTreesTmpTextLen > maxChildTreesTextLen:
                            maxChildTrees = childTreesTmp[i:j]
                            maxChildTreesTextLen = childTreesTmpTextLen
            # Compare html fragment of current tag branch and the ones of children:
            if maxChildSubTreesTextLen > maxChildTreesTextLen:
                maxChildTrees, maxChildTreesTextLen = maxChildSubTrees, maxChildSubTreesTextLen
    return (maxTree, maxTextLen, maxChildTrees, maxChildTreesTextLen)

def _filterSpam(densDic, threshold):
    """
    Walk through html document, drop off all etree branches that having low text
    density, and return the left parts of fragments.

    Return: html fragments of main text
    """
    dens, textLen, totalLen, tree = densDic['self']
    # If density is larger than threshold, keep and return current tag branch:
    if dens >= threshold:
        return etree.tostring(tree, encoding = unicode)
    if str(tree.tag).lower() == 'br':
        return etree.tostring(tree, encoding = unicode)
    # If density of current tag branch is too small, check its children:
    else:
        frags = []
        if densDic.has_key('child'):
            for childDic in densDic['child']:
                frags.append(_filterSpam(childDic, threshold))
        return ''.join(frags)

def _calcDensity(tree):
    """
    Calculate the text density for every etree branch. The define of text density is:
    (the length of pure text content under current html tag) / (total length of all content under current html tag)

    Return: {'self': (tag density, length of pure text, total length of html tags and text, etree instance), 
    'child': list of density dics for child entities }
    """
    tag = str(tree.tag).lower()
    if (tag == '<built-in function comment>' or
       tag == 'script' or
       tag == 'noscript' or
       tag == 'style'):
        return {'self': (0.0, 0, 0, tree)}
    text = tree.text if tree.text != None else ''
    tail = tree.tail if tree.tail != None else ''
    countTextLen = len(text.strip()) + len(tail.strip())
    totalLen = len(etree.tostring(tree, encoding = unicode)) if tree != None else 0
    if str(tree.tag).lower() == 'br':
        return {'self': (1.0 / totalLen, 1, totalLen, tree)}
    dicList = []
    treeOrig = tree[:]
    for subtree in treeOrig:
        textNode = None
        if subtree.tail and len(subtree.tail.strip()) > 0:
            index = tree.index(subtree)
            textNode = subtree.tail
            subtree.tail = ''
        dic = _calcDensity(subtree)
        dicList.append(dic)
        textLen = dic['self'][1]
        countTextLen += textLen
        # Treat subtree.tail as an independent etree branch:
        if textNode != None:
            textNodeTotalLen = len(textNode)
            textNodeTextLen = len(textNode.strip())
            textTree = etree.Element('span')
            textTree.text = textNode
            tree.insert(index + 1, textTree)
            dicList.append({'self': (float(textNodeTextLen) / textNodeTotalLen, textNodeTextLen, textNodeTotalLen, textTree)})
            countTextLen += textNodeTextLen
    density = float(countTextLen) / totalLen if totalLen != 0 else 0.0
    return {'self': (density, countTextLen, totalLen, tree), 'child': dicList}

if __name__ == '__main__':
    f = open('163.html', 'r')
    html = f.read()
    f.close()
    html = unicode(html, 'gb18030')
    # Get main text via 'Extract' mode (Get only main content of article):
#    mtHtml = extMainText(html)
#    print mtHtml
    # Get main text via 'Filter' mode (Most possibly get both main content & comments):
    mtHtml = extMainText(html, filterMode = True)
    print mtHtml
    
